@@ -1,6 +1,5 @@
 //! Core namespace for zizmor's audits.
 
-use line_index::LineIndex;
 use thiserror::Error;
 use tracing::instrument;
 use yamlpath::Document;
@@ -11,6 +10,7 @@ use crate::{
     models::{
         AsDocument,
         action::{Action, CompositeStep},
+        dependabot::Dependabot,
         workflow::{Job, NormalJob, ReusableWorkflowCallJob, Step, Workflow},
     },
     registry::input::InputKey,
@@ -22,6 +22,8 @@ pub(crate) mod artipacked;
 pub(crate) mod bot_conditions;
 pub(crate) mod cache_poisoning;
 pub(crate) mod dangerous_triggers;
+pub(crate) mod dependabot_cooldown;
+pub(crate) mod dependabot_execution;
 pub(crate) mod excessive_permissions;
 pub(crate) mod forbidden_uses;
 pub(crate) mod github_env;
@@ -32,6 +34,7 @@ pub(crate) mod known_vulnerable_actions;
 pub(crate) mod obfuscation;
 pub(crate) mod overprovisioned_secrets;
 pub(crate) mod ref_confusion;
+pub(crate) mod ref_version_mismatch;
 pub(crate) mod secrets_inherit;
 pub(crate) mod self_hosted_runner;
 pub(crate) mod stale_action_refs;
@@ -48,6 +51,7 @@ pub(crate) mod use_trusted_publishing;
 pub(crate) enum AuditInput {
     Workflow(Workflow),
     Action(Action),
+    Dependabot(Dependabot),
 }
 
 impl AuditInput {
@@ -55,13 +59,7 @@ impl AuditInput {
         match self {
             AuditInput::Workflow(workflow) => &workflow.key,
             AuditInput::Action(action) => &action.key,
-        }
-    }
-
-    pub(crate) fn line_index(&self) -> &LineIndex {
-        match self {
-            AuditInput::Workflow(workflow) => workflow.as_document().line_index(),
-            AuditInput::Action(action) => action.as_document().line_index(),
+            AuditInput::Dependabot(dependabot) => &dependabot.key,
         }
     }
 
@@ -69,6 +67,7 @@ impl AuditInput {
         match self {
             AuditInput::Workflow(workflow) => workflow.link.as_deref(),
             AuditInput::Action(action) => action.link.as_deref(),
+            AuditInput::Dependabot(dependabot) => dependabot.link.as_deref(),
         }
     }
 
@@ -76,6 +75,7 @@ impl AuditInput {
         match self {
             AuditInput::Workflow(workflow) => workflow.location(),
             AuditInput::Action(action) => action.location(),
+            AuditInput::Dependabot(dependabot) => dependabot.location(),
         }
     }
 }
@@ -85,6 +85,7 @@ impl<'a> AsDocument<'a, 'a> for AuditInput {
         match self {
             AuditInput::Workflow(workflow) => workflow.as_document(),
             AuditInput::Action(action) => action.as_document(),
+            AuditInput::Dependabot(dependabot) => dependabot.as_document(),
         }
     }
 }
@@ -101,6 +102,12 @@ impl From<Action> for AuditInput {
     }
 }
 
+impl From<Dependabot> for AuditInput {
+    fn from(value: Dependabot) -> Self {
+        Self::Dependabot(value)
+    }
+}
+
 /// A supertrait for all audits.
 ///
 /// Workflow audits, action audits, and all future audit types
@@ -277,6 +284,14 @@ pub(crate) trait Audit: AuditCore {
         Ok(results)
     }
 
+    fn audit_dependabot<'doc>(
+        &self,
+        _dependabot: &'doc Dependabot,
+        _config: &Config,
+    ) -> anyhow::Result<Vec<Finding<'doc>>> {
+        Ok(vec![])
+    }
+
     fn audit_raw<'doc>(
         &self,
         _input: &'doc AuditInput,
@@ -315,6 +330,7 @@ pub(crate) trait Audit: AuditCore {
         let mut results = match input {
             AuditInput::Workflow(workflow) => self.audit_workflow(workflow, config),
             AuditInput::Action(action) => self.audit_action(action, config),
+            AuditInput::Dependabot(dependabot) => self.audit_dependabot(dependabot, config),
         }?;
 
         results.extend(self.audit_raw(input, config)?);
