@@ -1,32 +1,37 @@
 #![warn(clippy::all, clippy::dbg_macro)]
 
 use std::{
+    collections::HashSet,
     io::{Write, stdout},
     process::ExitCode,
-    str::FromStr,
 };
 
-use annotate_snippets::{Level, Renderer};
+use annotate_snippets::{Group, Level, Renderer};
 use anstream::{eprintln, println, stream::IsTerminal};
-use anyhow::{Context, Result, anyhow};
-use camino::Utf8Path;
+use anyhow::anyhow;
+use camino::Utf8PathBuf;
 use clap::{Args, CommandFactory, Parser, ValueEnum, builder::NonEmptyStringValueParser};
 use clap_complete::Generator;
 use clap_verbosity_flag::InfoLevel;
-use config::Config;
+use etcetera::AppStrategy as _;
 use finding::{Confidence, Persona, Severity};
 use github_api::{GitHubHost, GitHubToken};
-use ignore::WalkBuilder;
 use indicatif::ProgressStyle;
 use owo_colors::OwoColorize;
-use registry::{AuditRegistry, FindingRegistry, InputKey, InputKind, InputRegistry};
+use registry::input::{InputKey, InputRegistry};
+use registry::{AuditRegistry, FindingRegistry};
 use state::AuditState;
 use terminal_link::Link;
-use tracing::{Span, info_span, instrument};
+use thiserror::Error;
+use tracing::{Span, info_span, instrument, warn};
 use tracing_indicatif::{IndicatifLayer, span_ext::IndicatifSpanExt};
 use tracing_subscriber::{EnvFilter, layer::SubscriberExt as _, util::SubscriberInitExt as _};
 
-use crate::registry::RepoSlug;
+use crate::{
+    config::{Config, ConfigError, ConfigErrorInner},
+    github_api::Client,
+    registry::input::CollectionError,
+};
 
 mod audit;
 mod config;
@@ -40,6 +45,18 @@ mod registry;
 mod state;
 mod utils;
 
+#[cfg(all(
+    not(target_env = "msvc"),
+    any(
+        target_arch = "x86_64",
+        target_arch = "aarch64",
+        // NOTE(ww): Not a build we currently support.
+        // target_arch = "powerpc64"
+    )
+))]
+#[global_allocator]
+static GLOBAL: tikv_jemallocator::Jemalloc = tikv_jemallocator::Jemalloc;
+
 // TODO: Dedupe this with the top-level `sponsors.json` used by the
 // README + docs site.
 const THANKS: &[(&str, &str)] = &[("Grafana Labs", "https://grafana.com")];
@@ -74,7 +91,7 @@ struct App {
     gh_token: Option<GitHubToken>,
 
     /// The GitHub Server Hostname. Defaults to github.com
-    #[arg(long, env = "GH_HOST", default_value = "github.com", value_parser = GitHubHost::new)]
+    #[arg(long, env = "GH_HOST", default_value_t)]
     gh_hostname: GitHubHost,
 
     /// Perform only offline audits.
@@ -100,8 +117,9 @@ struct App {
     #[arg(long, value_enum, value_name = "MODE")]
     color: Option<ColorMode>,
 
-    /// The configuration file to load. By default, any config will be
-    /// discovered relative to $CWD.
+    /// The configuration file to load.
+    /// This loads a single configuration file across all input groups,
+    /// which may not be what you intend.
     #[arg(
         short,
         long,
@@ -121,23 +139,23 @@ struct App {
 
     /// Filter all results below this severity.
     #[arg(long)]
-    min_severity: Option<Severity>,
+    min_severity: Option<CliSeverity>,
 
     /// Filter all results below this confidence.
     #[arg(long)]
-    min_confidence: Option<Confidence>,
+    min_confidence: Option<CliConfidence>,
 
     /// The directory to use for HTTP caching. By default, a
     /// host-appropriate user-caching directory will be used.
-    #[arg(long)]
-    cache_dir: Option<String>,
+    #[arg(long, default_value_t = App::default_cache_dir(), hide_default_value = true)]
+    cache_dir: Utf8PathBuf,
 
     /// Control which kinds of inputs are collected for auditing.
     ///
     /// By default, all workflows and composite actions are collected,
     /// while honoring `.gitignore` files.
-    #[arg(long, value_enum, default_value_t)]
-    collect: CollectionMode,
+    #[arg(long, default_values = ["default"], num_args=1.., value_delimiter=',')]
+    collect: Vec<CliCollectionMode>,
 
     /// Fail instead of warning on syntax and schema errors
     /// in collected inputs.
@@ -180,6 +198,43 @@ struct App {
     inputs: Vec<String>,
 }
 
+impl App {
+    fn default_cache_dir() -> Utf8PathBuf {
+        etcetera::choose_app_strategy(etcetera::AppStrategyArgs {
+            top_level_domain: "io.github".into(),
+            author: "woodruffw".into(),
+            app_name: "zizmor".into(),
+        })
+        .expect("failed to determine default cache directory")
+        .cache_dir()
+        .try_into()
+        .expect("failed to turn cache directory into a sane path")
+    }
+}
+
+// NOTE(ww): This can be removed once `--min-severity=unknown`
+// is fully removed.
+#[derive(Debug, Copy, Clone, ValueEnum)]
+enum CliSeverity {
+    #[value(hide = true)]
+    Unknown,
+    Informational,
+    Low,
+    Medium,
+    High,
+}
+
+// NOTE(ww): This can be removed once `--min-confidence=unknown`
+// is fully removed.
+#[derive(Debug, Copy, Clone, ValueEnum)]
+enum CliConfidence {
+    #[value(hide = true)]
+    Unknown,
+    Low,
+    Medium,
+    High,
+}
+
 #[cfg(feature = "lsp")]
 #[derive(Args)]
 #[group(multiple = true, conflicts_with = "inputs")]
@@ -301,39 +356,128 @@ impl From<ColorMode> for anstream::ColorChoice {
 }
 
 /// How `zizmor` collects inputs from local and remote repository sources.
-#[derive(Copy, Clone, Debug, Default, ValueEnum)]
-pub(crate) enum CollectionMode {
+#[derive(Copy, Clone, Debug, Default, ValueEnum, Eq, PartialEq, Hash)]
+pub(crate) enum CliCollectionMode {
     /// Collect all possible inputs, ignoring `.gitignore` files.
     All,
     /// Collect all possible inputs, respecting `.gitignore` files.
     #[default]
     Default,
     /// Collect only workflow definitions.
+    ///
+    /// Deprecated; use `--collect=workflows`
+    #[value(hide = true)]
     WorkflowsOnly,
     /// Collect only action definitions (i.e. `action.yml`).
+    ///
+    /// Deprecated; use `--collect=actions`
+    #[value(hide = true)]
     ActionsOnly,
+    /// Collect workflows.
+    Workflows,
+    /// Collect action definitions (i.e. `action.yml`).
+    Actions,
+    /// Collect Dependabot configuration files (i.e. `dependabot.yml`).
+    Dependabot,
 }
 
-impl CollectionMode {
+#[derive(Copy, Clone, Debug, Eq, PartialEq, Hash)]
+pub(crate) enum CollectionMode {
+    All,
+    Default,
+    Workflows,
+    Actions,
+    Dependabot,
+}
+
+pub(crate) struct CollectionModeSet(HashSet<CollectionMode>);
+
+impl From<&[CliCollectionMode]> for CollectionModeSet {
+    fn from(modes: &[CliCollectionMode]) -> Self {
+        if modes.len() > 1
+            && modes.iter().any(|mode| {
+                matches!(
+                    mode,
+                    CliCollectionMode::WorkflowsOnly | CliCollectionMode::ActionsOnly
+                )
+            })
+        {
+            let mut cmd = App::command();
+
+            cmd.error(
+                clap::error::ErrorKind::ArgumentConflict,
+                "`workflows-only` and `actions-only` cannot be combined with other collection modes",
+            )
+            .exit();
+        }
+
+        Self(
+            modes
+                .iter()
+                .map(|mode| match mode {
+                    CliCollectionMode::All => CollectionMode::All,
+                    CliCollectionMode::Default => CollectionMode::Default,
+                    CliCollectionMode::WorkflowsOnly => {
+                        warn!("--collect=workflows-only is deprecated; use --collect=workflows instead");
+                        warn!("future versions of zizmor will reject this mode");
+
+                        CollectionMode::Workflows
+                    }
+                    CliCollectionMode::ActionsOnly => {
+                        warn!("--collect=actions-only is deprecated; use --collect=actions instead");
+                        warn!("future versions of zizmor will reject this mode");
+
+                        CollectionMode::Actions
+                    }
+                    CliCollectionMode::Workflows => CollectionMode::Workflows,
+                    CliCollectionMode::Actions => CollectionMode::Actions,
+                    CliCollectionMode::Dependabot => CollectionMode::Dependabot,
+                })
+                .collect(),
+        )
+    }
+}
+
+impl CollectionModeSet {
+    /// Does our collection mode respect `.gitignore` files?
     pub(crate) fn respects_gitignore(&self) -> bool {
-        matches!(
-            self,
-            CollectionMode::Default | CollectionMode::WorkflowsOnly | CollectionMode::ActionsOnly
-        )
+        // All modes except 'all' respect .gitignore files.
+        !self.0.contains(&CollectionMode::All)
     }
 
+    /// Should we collect workflows?
     pub(crate) fn workflows(&self) -> bool {
-        matches!(
-            self,
-            CollectionMode::All | CollectionMode::Default | CollectionMode::WorkflowsOnly
-        )
+        self.0.iter().any(|mode| {
+            matches!(
+                mode,
+                CollectionMode::All | CollectionMode::Default | CollectionMode::Workflows
+            )
+        })
     }
 
+    /// Should we collect *only* workflows?
+    pub(crate) fn workflows_only(&self) -> bool {
+        self.0.len() == 1 && self.0.contains(&CollectionMode::Workflows)
+    }
+
+    /// Should we collect actions?
     pub(crate) fn actions(&self) -> bool {
-        matches!(
-            self,
-            CollectionMode::All | CollectionMode::Default | CollectionMode::ActionsOnly
-        )
+        self.0.iter().any(|mode| {
+            matches!(
+                mode,
+                CollectionMode::All | CollectionMode::Default | CollectionMode::Actions
+            )
+        })
+    }
+
+    /// Should we collect Dependabot configuration files?
+    pub(crate) fn dependabot(&self) -> bool {
+        self.0.iter().any(|mode| {
+            matches!(
+                mode,
+                CollectionMode::All | CollectionMode::Default | CollectionMode::Dependabot
+            )
+        })
     }
 }
 
@@ -348,177 +492,40 @@ pub(crate) enum FixMode {
 }
 
 pub(crate) fn tips(err: impl AsRef<str>, tips: &[impl AsRef<str>]) -> String {
-    let mut message = Level::Error.title(err.as_ref());
-    for tip in tips {
-        message = message.footer(Level::Note.title(tip.as_ref()));
-    }
+    // NOTE: We use secondary_title here because primary_title doesn't
+    // allow ANSI colors, and some of our errors contain colorized text.
+    let report = vec![
+        Group::with_title(Level::ERROR.secondary_title(err.as_ref()))
+            .elements(tips.iter().map(|tip| Level::HELP.message(tip.as_ref()))),
+    ];
 
     let renderer = Renderer::styled();
-    format!("{}", renderer.render(message))
+    renderer.render(&report).to_string()
 }
 
-#[instrument(skip(mode, registry))]
-fn collect_from_dir(
-    input_path: &Utf8Path,
-    mode: &CollectionMode,
-    registry: &mut InputRegistry,
-) -> Result<()> {
-    // Start with all filters disabled, i.e. walk everything.
-    let mut walker = WalkBuilder::new(input_path);
-    let walker = walker.standard_filters(false);
-
-    // If the user wants to respect `.gitignore` files, then we need to
-    // explicitly enable it. This also enables filtering by a global
-    // `.gitignore` file and the `.git/info/exclude` file, since these
-    // typically align with the user's expectations.
-    //
-    // We honor `.gitignore` and similar files even if `.git/` is not
-    // present, since users may retrieve or reconstruct a source archive
-    // without a `.git/` directory. In particular, this snares some
-    // zizmor integrators.
-    //
-    // See: https://github.com/zizmorcore/zizmor/issues/596
-    if mode.respects_gitignore() {
-        walker
-            .require_git(false)
-            .git_ignore(true)
-            .git_global(true)
-            .git_exclude(true);
-    }
-
-    for entry in walker.build() {
-        let entry = entry?;
-        let entry = <&Utf8Path>::try_from(entry.path())?;
-
-        if mode.workflows()
-            && entry.is_file()
-            && matches!(entry.extension(), Some("yml" | "yaml"))
-            && entry
-                .parent()
-                .is_some_and(|dir| dir.ends_with(".github/workflows"))
-        {
-            let key = InputKey::local(entry, Some(input_path))?;
-            let contents = std::fs::read_to_string(entry)?;
-            registry.register(InputKind::Workflow, contents, key)?;
-        }
-
-        if mode.actions()
-            && entry.is_file()
-            && matches!(entry.file_name(), Some("action.yml" | "action.yaml"))
-        {
-            let key = InputKey::local(entry, Some(input_path))?;
-            let contents = std::fs::read_to_string(entry)?;
-            registry.register(InputKind::Action, contents, key)?;
-        }
-    }
-
-    Ok(())
-}
-
-fn collect_from_repo_slug(
-    input: &str,
-    mode: &CollectionMode,
-    state: &AuditState,
-    registry: &mut InputRegistry,
-) -> Result<()> {
-    let Ok(slug) = RepoSlug::from_str(input) else {
-        return Err(anyhow!(tips(
-            format!("invalid input: {input}"),
-            &[format!(
-                "pass a single {file}, {directory}, or entire repo by {slug} slug",
-                file = "file".green(),
-                directory = "directory".green(),
-                slug = "owner/repo".green()
-            )]
-        )));
-    };
-
-    let client = state.gh_client.as_ref().ok_or_else(|| {
-        anyhow!(tips(
-            format!("can't retrieve repository: {input}", input = input.green()),
-            &[format!(
-                "try removing {offline} or passing {gh_token}",
-                offline = "--offline".yellow(),
-                gh_token = "--gh-token <TOKEN>".yellow(),
-            )]
-        ))
-    })?;
-
-    if matches!(mode, CollectionMode::WorkflowsOnly) {
-        // Performance: if we're *only* collecting workflows, then we
-        // can save ourselves a full repo download and only fetch the
-        // repo's workflow files.
-        client.fetch_workflows(&slug, registry)?;
-    } else {
-        let before = registry.len();
-        let host = match &state.gh_hostname {
-            GitHubHost::Enterprise(address) => address.as_str(),
-            GitHubHost::Standard(_) => "github.com",
-        };
-
-        client
-            .fetch_audit_inputs(&slug, registry)
-            .with_context(|| {
-                tips(
-                    format!(
-                        "couldn't collect inputs from https://{host}/{owner}/{repo}",
-                        host = host,
-                        owner = slug.owner,
-                        repo = slug.repo
-                    ),
-                    &["confirm the repository exists and that you have access to it"],
-                )
-            })?;
-        let after = registry.len();
-        let len = after - before;
-
-        tracing::info!(
-            "collected {len} inputs from {owner}/{repo}",
-            owner = slug.owner,
-            repo = slug.repo
-        );
-    }
-
-    Ok(())
+/// State used when collecting input groups.
+pub(crate) struct CollectionOptions {
+    pub(crate) mode_set: CollectionModeSet,
+    pub(crate) strict: bool,
+    pub(crate) no_config: bool,
+    /// Global configuration, if any.
+    pub(crate) global_config: Option<Config>,
 }
 
 #[instrument(skip_all)]
 fn collect_inputs(
     inputs: &[String],
-    mode: &CollectionMode,
-    strict: bool,
-    state: &AuditState,
-) -> Result<InputRegistry> {
-    let mut registry = InputRegistry::new(strict);
+    options: &CollectionOptions,
+    gh_client: Option<&Client>,
+) -> Result<InputRegistry, CollectionError> {
+    let mut registry = InputRegistry::new();
 
-    for input in inputs {
-        let input_path = Utf8Path::new(input);
-        if input_path.is_file() {
-            // When collecting individual files, we don't know which part
-            // of the input path is the prefix.
-            let (key, kind) = match (input_path.file_stem(), input_path.extension()) {
-                (Some("action"), Some("yml" | "yaml")) => {
-                    (InputKey::local(input_path, None)?, InputKind::Action)
-                }
-                (Some(_), Some("yml" | "yaml")) => {
-                    (InputKey::local(input_path, None)?, InputKind::Workflow)
-                }
-                _ => return Err(anyhow!("invalid input: {input}")),
-            };
-
-            let contents = std::fs::read_to_string(input_path)?;
-            registry.register(kind, contents, key)?;
-        } else if input_path.is_dir() {
-            collect_from_dir(input_path, mode, &mut registry)?;
-        } else {
-            // If this input isn't a file or directory, it's probably an
-            // `owner/repo(@ref)?` slug.
-            collect_from_repo_slug(input, mode, state, &mut registry)?;
-        }
+    for input in inputs.iter() {
+        registry.register_group(input, options, gh_client)?;
     }
 
     if registry.len() == 0 {
-        return Err(anyhow!("no inputs collected"));
+        return Err(CollectionError::NoInputs);
     }
 
     Ok(registry)
@@ -533,11 +540,40 @@ fn completions<G: clap_complete::Generator>(generator: G, cmd: &mut clap::Comman
     );
 }
 
-fn run() -> Result<ExitCode> {
-    human_panic::setup_panic!();
-
-    let mut app = App::parse();
+/// Top-level errors.
+#[derive(Debug, Error)]
+enum Error {
+    /// An error in global configuration.
+    #[error(transparent)]
+    GlobalConfig(#[from] ConfigError),
+    /// An error while collecting inputs.
+    #[error(transparent)]
+    Collection(#[from] CollectionError),
+    /// An error while running the LSP server.
+    #[error(transparent)]
+    Lsp(#[from] lsp::Error),
+    /// An error from the GitHub API client.
+    #[error(transparent)]
+    Client(#[from] github_api::ClientError),
+    /// An error while loading audit rules.
+    #[error("failed to load audit rules")]
+    AuditLoad(#[source] anyhow::Error),
+    /// An error while running an audit.
+    #[error("{ident} failed on {input}")]
+    Audit {
+        source: anyhow::Error,
+        ident: &'static str,
+        input: String,
+    },
+    /// An error while rendering output.
+    #[error("failed to render output")]
+    Output(#[source] anyhow::Error),
+    /// An error while performing fixes.
+    #[error("failed to apply fixes")]
+    Fix(#[source] anyhow::Error),
+}
 
+fn run(app: &mut App) -> Result<ExitCode, Error> {
     #[cfg(feature = "lsp")]
     if app.lsp.lsp {
         lsp::run()?;
@@ -610,7 +646,8 @@ fn run() -> Result<ExitCode> {
 
     let filter = EnvFilter::builder()
         .with_default_directive(app.verbose.tracing_level_filter().into())
-        .from_env()?;
+        .from_env()
+        .expect("failed to parse RUST_LOG");
 
     let reg = tracing_subscriber::registry()
         .with(
@@ -628,28 +665,61 @@ fn run() -> Result<ExitCode> {
         reg.with(indicatif_layer).init();
     }
 
-    let config = Config::new(&app).map_err(|e| {
-        anyhow!(tips(
-            format!("failed to load config: {e:#}"),
-            &[
-                "check your configuration file for errors",
-                "see: https://docs.zizmor.sh/configuration/"
-            ]
-        ))
-    })?;
+    eprintln!("🌈 zizmor v{version}", version = env!("CARGO_PKG_VERSION"));
+
+    let collection_mode_set = CollectionModeSet::from(app.collect.as_slice());
+
+    let min_severity = match app.min_severity {
+        Some(CliSeverity::Unknown) => {
+            tracing::warn!("`unknown` is a deprecated minimum severity that has no effect");
+            tracing::warn!("future versions of zizmor will reject this value");
+            None
+        }
+        Some(CliSeverity::Informational) => Some(Severity::Informational),
+        Some(CliSeverity::Low) => Some(Severity::Low),
+        Some(CliSeverity::Medium) => Some(Severity::Medium),
+        Some(CliSeverity::High) => Some(Severity::High),
+        None => None,
+    };
+
+    let min_confidence = match app.min_confidence {
+        Some(CliConfidence::Unknown) => {
+            tracing::warn!("`unknown` is a deprecated minimum confidence that has no effect");
+            tracing::warn!("future versions of zizmor will reject this value");
+            None
+        }
+        Some(CliConfidence::Low) => Some(Confidence::Low),
+        Some(CliConfidence::Medium) => Some(Confidence::Medium),
+        Some(CliConfidence::High) => Some(Confidence::High),
+        None => None,
+    };
+
+    let global_config = Config::global(app)?;
+
+    let gh_client = app
+        .gh_token
+        .as_ref()
+        .map(|token| Client::new(&app.gh_hostname, token, &app.cache_dir))
+        .transpose()?;
+
+    let collection_options = CollectionOptions {
+        mode_set: collection_mode_set,
+        strict: app.strict_collection,
+        no_config: app.no_config,
+        global_config,
+    };
 
-    let audit_state = AuditState::new(&app, &config)?;
     let registry = collect_inputs(
-        &app.inputs,
-        &app.collect,
-        app.strict_collection,
-        &audit_state,
+        app.inputs.as_slice(),
+        &collection_options,
+        gh_client.as_ref(),
     )?;
 
-    let audit_registry = AuditRegistry::default_audits(&audit_state)?;
+    let state = AuditState::new(app.no_online_audits, gh_client);
 
-    let mut results =
-        FindingRegistry::new(app.min_severity, app.min_confidence, app.persona, &config);
+    let audit_registry = AuditRegistry::default_audits(&state).map_err(Error::AuditLoad)?;
+
+    let mut results = FindingRegistry::new(&registry, min_severity, min_confidence, app.persona);
     {
         // Note: block here so that we drop the span here at the right time.
         let span = info_span!("audit");
@@ -660,40 +730,47 @@ fn run() -> Result<ExitCode> {
 
         let _guard = span.enter();
 
-        for (_, input) in registry.iter_inputs() {
+        for (input_key, input) in registry.iter_inputs() {
             Span::current().pb_set_message(input.key().filename());
-            for (name, audit) in audit_registry.iter_audits() {
-                tracing::debug!(
-                    "running {name} on {input}",
-                    name = name,
-                    input = input.key()
+            let config = registry.get_config(input_key.group());
+            for (ident, audit) in audit_registry.iter_audits() {
+                tracing::debug!("running {ident} on {input}", input = input.key());
+
+                results.extend(
+                    audit
+                        .audit(ident, input, config)
+                        .map_err(|err| Error::Audit {
+                            source: err,
+                            ident,
+                            input: input.key().to_string(),
+                        })?,
                 );
-                results.extend(audit.audit(input).with_context(|| {
-                    format!("{name} failed on {input}", input = input.key().filename())
-                })?);
+
                 Span::current().pb_inc(1);
             }
             tracing::info!(
-                "🌈 {completed} {input}",
-                completed = "completed".green(),
+                "🌈 completed {input}",
                 input = input.key().presentation_path()
             );
         }
     }
 
     match app.format {
-        OutputFormat::Plain => output::plain::render_findings(&app, &registry, &results),
+        OutputFormat::Plain => output::plain::render_findings(&registry, &results, app.naches),
         OutputFormat::Json | OutputFormat::JsonV1 => {
-            output::json::v1::output(stdout(), results.findings())?
+            output::json::v1::output(stdout(), results.findings()).map_err(Error::Output)?
         }
         OutputFormat::Sarif => {
-            serde_json::to_writer_pretty(stdout(), &output::sarif::build(results.findings()))?
+            serde_json::to_writer_pretty(stdout(), &output::sarif::build(results.findings()))
+                .map_err(|err| Error::Output(anyhow!(err)))?
+        }
+        OutputFormat::Github => {
+            output::github::output(stdout(), results.findings()).map_err(Error::Output)?
         }
-        OutputFormat::Github => output::github::output(stdout(), results.findings())?,
     };
 
     if let Some(fix_mode) = app.fix {
-        output::fix::apply_fixes(fix_mode, &results, &registry)?;
+        output::fix::apply_fixes(fix_mode, &results, &registry).map_err(Error::Fix)?;
     }
 
     if app.no_exit_codes || matches!(app.format, OutputFormat::Sarif) {
@@ -704,15 +781,109 @@ fn run() -> Result<ExitCode> {
 }
 
 fn main() -> ExitCode {
+    human_panic::setup_panic!();
+
+    let mut app = App::parse();
+
     // This is a little silly, but returning an ExitCode like this ensures
     // we always exit cleanly, rather than performing a hard process exit.
-    match run() {
+    match run(&mut app) {
         Ok(exit) => exit,
         Err(err) => {
             eprintln!(
                 "{fatal}: no audit was performed",
                 fatal = "fatal".red().bold()
             );
+
+            let report = match &err {
+                // NOTE(ww): Slightly annoying that we have two different config error
+                // wrapper states, but oh well.
+                Error::GlobalConfig(err) | Error::Collection(CollectionError::Config(err)) => {
+                    let mut group = Group::with_title(Level::ERROR.primary_title(err.to_string()));
+
+                    match err.source {
+                        ConfigErrorInner::Syntax(_) => {
+                            group = group.elements([
+                                Level::HELP
+                                    .message("check your configuration file for syntax errors"),
+                                Level::HELP.message("see: https://docs.zizmor.sh/configuration/"),
+                            ]);
+                        }
+                        ConfigErrorInner::AuditSyntax(_, ident) => {
+                            group = group.elements([
+                                Level::HELP.message(format!(
+                                    "check the configuration for the '{ident}' rule"
+                                )),
+                                Level::HELP.message(format!(
+                                    "see: https://docs.zizmor.sh/audits/#{ident}-configuration"
+                                )),
+                            ]);
+                        }
+                        _ => {}
+                    }
+
+                    let renderer = Renderer::styled();
+                    let report = renderer.render(&[group]);
+
+                    Some(report)
+                }
+                Error::Collection(err @ CollectionError::InvalidInput(..)) => {
+                    let group = Group::with_title(Level::ERROR.primary_title(err.to_string()))
+                        .element(Level::HELP.message(format!(
+                            "valid inputs are files, directories, or GitHub {slug} slugs",
+                            slug = "user/repo[@ref]".green()
+                        )))
+                        .element(Level::HELP.message(format!(
+                            "examples: {ex1}, {ex2}, {ex3}, or {ex4}",
+                            ex1 = "path/to/workflow.yml".green(),
+                            ex2 = ".github/".green(),
+                            ex3 = "example/example".green(),
+                            ex4 = "example/example@v1.2.3".green()
+                        )));
+
+                    let renderer = Renderer::styled();
+                    let report = renderer.render(&[group]);
+
+                    Some(report)
+                }
+                Error::Collection(err @ CollectionError::NoGitHubClient(_)) => {
+                    let mut group = Group::with_title(Level::ERROR.primary_title(err.to_string()));
+
+                    if app.offline {
+                        group = group
+                            .elements([Level::HELP
+                                .message("remove --offline to audit remote repositories")]);
+                    } else if app.gh_token.is_none() {
+                        group = group
+                            .elements([Level::HELP
+                                .message("set a GitHub token with --gh-token or GH_TOKEN")]);
+                    }
+
+                    let renderer = Renderer::styled();
+                    let report = renderer.render(&[group]);
+
+                    Some(report)
+                }
+                Error::Collection(err @ CollectionError::Yamlpath(_)) => {
+                    let group = Group::with_title(Level::ERROR.primary_title(err.to_string())).elements([
+                        Level::HELP.message("this typically indicates a bug in zizmor; please report it"),
+                        Level::HELP.message(
+                            "https://github.com/zizmorcore/zizmor/issues/new?template=bug-report.yml",
+                        ),
+                    ]);
+                    let renderer = Renderer::styled();
+                    let report = renderer.render(&[group]);
+
+                    Some(report)
+                }
+                _ => None,
+            };
+
+            let mut err = anyhow!(err);
+            if let Some(report) = report {
+                err = err.context(report);
+            }
+
             eprintln!("{err:?}");
             ExitCode::FAILURE
         }
