@@ -4,6 +4,7 @@ use anyhow::Result;
 
 use crate::common::{OutputMode, input_under_test, zizmor};
 
+mod collect;
 mod json_v1;
 
 #[cfg_attr(not(feature = "gh-token-tests"), ignore)]
@@ -33,7 +34,7 @@ fn issue_569() -> Result<()> {
         zizmor()
             .offline(false)
             .output(OutputMode::Both)
-            .args(["--no-online-audits", "--collect=workflows-only"])
+            .args(["--no-online-audits", "--collect=workflows"])
             .input("python/cpython@f963239ff1f986742d4c6bab2ab7b73f5a4047f6")
             .run()?
     );
@@ -344,3 +345,24 @@ fn warn_on_min_confidence_unknown() -> Result<()> {
     );
     Ok(())
 }
+
+/// Regression test for #1207.
+///
+/// Ensures that we correctly handle single-inputs that aren't given
+/// with an explicit parent path, e.g. `action.yml` instead of
+/// `./action.yml`.
+#[test]
+fn issue_1207() -> Result<()> {
+    insta::assert_snapshot!(
+        zizmor()
+            .expects_failure(false)
+            .output(OutputMode::Both)
+            .working_dir(input_under_test("e2e-menagerie/dummy-action-1"))
+            // Input doesn't matter, as long as it's relative without a leading
+            // `./` or other path component.
+            .input("action.yaml")
+            .run()?
+    );
+
+    Ok(())
+}
